# -*- coding: utf-8 -*-
"""Cursor handling utilities."""

# Third Party Libraries
from qtpy.QtCore import QTimer
from qtpy.QtGui import QBrush, QColor, QTextBlock, QTextCursor
from qtpy.QtWidgets import QTextEdit
from spyder.config.manager import CONF
from spyder.plugins.editor.api.decoration import DRAW_ORDERS

# Project Libraries
from spyder_okvim.spyder.config import CONF_SECTION
from spyder_okvim.utils.motion import MotionInfo, MotionType


class VimCursor:
    """Manage the Vim cursor."""

    def __init__(self, editor_widget):
        self.editor_widget = editor_widget

        self.vim_cursor = QTextEdit.ExtraSelection()
        self.vim_cursor.format.setForeground(QBrush(QColor("#000000")))
        self.vim_cursor.format.setBackground(QBrush(QColor("#BBBBBB")))

        self.selection = QTextEdit.ExtraSelection()
        self.selection.format.setForeground(QBrush(QColor("#A9B7C6")))
        self.selection.format.setBackground(QBrush(QColor("#214283")))

        self.yank_fg_color = QBrush(QColor("#B9C7D6"))
        self.yank_bg_color = QBrush(QColor("#7D7920"))
        self.hl_yank_dur = 400  # duration of highlight after yank
        self.hl_yank = True

        self.set_config_from_conf()

        # Order of Selections
        DRAW_ORDERS["vim_search"] = 6
        DRAW_ORDERS["hl_yank"] = 7
        DRAW_ORDERS["vim_selection"] = 8
        DRAW_ORDERS["vim_cursor"] = 9

    def set_config_from_conf(self):
        """Set config from conf."""
        self.vim_cursor.format.setForeground(
            QBrush(QColor(CONF.get(CONF_SECTION, "cursor_fg_color")))
        )
        self.vim_cursor.format.setBackground(
            QBrush(QColor(CONF.get(CONF_SECTION, "cursor_bg_color")))
        )

        self.selection.format.setForeground(
            QBrush(QColor(CONF.get(CONF_SECTION, "select_fg_color")))
        )
        self.selection.format.setBackground(
            QBrush(QColor(CONF.get(CONF_SECTION, "select_bg_color")))
        )

        self.yank_fg_color = QBrush(QColor(CONF.get(CONF_SECTION, "yank_fg_color")))
        self.yank_bg_color = QBrush(QColor(CONF.get(CONF_SECTION, "yank_bg_color")))
        self.hl_yank_dur = CONF.get(CONF_SECTION, "highlight_yank_duration")
        self.hl_yank = CONF.get(CONF_SECTION, "highlight_yank")

    def get_editor(self):
        """Get the editor focused."""
        editorstack = self.editor_widget.get_current_editorstack()
        return editorstack.get_current_editor()

    def get_editorstack(self):
        """Get the editorstack."""
        return self.editor_widget.get_current_editorstack()

    def get_cursor(self):
        """Get the cursor."""
        return self.get_editor().textCursor()

    def set_cursor(self, cursor):
        """Set the cursor to focused editor."""
        return self.get_editor().setTextCursor(cursor)

    def get_end_position(self):
        """Get the end position of document."""
        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.End)
        return cursor.position()

    def draw_vim_cursor(self):
        """Draw vim cursor."""
        vim_cursor = self.vim_cursor
        editor = self.get_editor()
        vim_cursor.cursor = editor.textCursor()
        vim_cursor.cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_cursor", [vim_cursor])

    def create_selection(self, start, end):
        """Create a text selection between two positions.

        Args:
            start: Starting position of the selection.
            end: End position of the selection.
        """
        end_document = self.get_end_position()
        if end > end_document:
            end = end_document

        sel = self.selection
        sel.cursor = QTextCursor(self.get_editor().document())
        sel.cursor.setPosition(start)
        sel.cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_selection", [sel])
        self.draw_vim_cursor()

    def set_cursor_pos_in_visual(self, pos_new):
        """Move the cursor while adjusting the current visual selection.

        Args:
            pos_new: Desired new cursor position.
        """
        if pos_new is None:
            return
        editor = self.get_editor()
        pos_cur = editor.textCursor().position()

        sel = editor.get_extra_selections("vim_selection")[0]
        start_old = sel.cursor.selectionStart()
        end_old = sel.cursor.selectionEnd()

        start = start_old
        end = end_old

        if abs(start_old - pos_cur) <= abs(end_old - 1 - pos_cur):
            if start_old >= pos_new:
                start = pos_new
            elif end_old > pos_new:
                start = pos_new
            else:
                start = end - 1
                end = pos_new + 1
        else:
            if end_old - 1 <= pos_new:
                end = pos_new + 1
            elif start_old < pos_new:
                end = pos_new + 1
            else:
                end = start_old + 1
                start = pos_new

        end_document = self.get_end_position()
        if end > end_document:
            end = end_document

        sel.cursor.setPosition(start)
        sel.cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_selection", [sel])
        self.set_cursor_pos(pos_new)

    def get_block(self, pos: int) -> tuple[QTextBlock, int]:
        """Get block number of cursor position."""
        cursor = self.get_cursor()
        cursor.setPosition(pos)
        return cursor.block(), cursor.blockNumber()

    def get_block_no_start_in_selection(self):
        """Get start block number of selection."""
        editor = self.get_editor()

        sel = editor.get_extra_selections("vim_selection")
        if not sel:
            return

        sel = sel[0]

        sel_start = sel.cursor.selectionStart()

        _, block_no_start = self.get_block(sel_start)

        return block_no_start

    def get_block_no_end_in_selection(self):
        """Get the last block number of selection."""
        editor = self.get_editor()

        sel = editor.get_extra_selections("vim_selection")
        if not sel:
            return

        sel = sel[0]

        sel_end = sel.cursor.selectionEnd()

        _, block_no_end = self.get_block(sel_end)

        return block_no_end

    def get_pos_start_in_selection(self):
        """Get start position of selection."""
        editor = self.get_editor()

        sel = editor.get_extra_selections("vim_selection")
        if not sel:
            return

        sel = sel[0]

        sel_start = sel.cursor.selectionStart()

        return sel_start

    def get_pos_end_in_selection(self):
        """Get end position of selection."""
        editor = self.get_editor()

        sel = editor.get_extra_selections("vim_selection")
        if not sel:
            return

        sel = sel[0]

        sel_end = sel.cursor.selectionEnd()

        return sel_end

    def set_selection_to_editor_using_vim_selection(self):
        """Set the selection to editor by referring the selection of vim."""
        editor = self.get_editor()
        old_cursor = editor.textCursor()
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()
        if sel_start:
            new_cursor = editor.textCursor()
            new_cursor.setPosition(sel_start, QTextCursor.MoveAnchor)
            new_cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
            editor.setTextCursor(new_cursor)
            return old_cursor

    def set_cursor_pos_in_vline(self, pos_new):
        """Move the cursor while updating the visual line selection.

        Args:
            pos_new: Desired new cursor position.
        """
        if pos_new is None:
            return
        editor = self.get_editor()
        block_no_cur = editor.textCursor().blockNumber()

        sel = editor.get_extra_selections("vim_selection")[0]
        sel_start = sel.cursor.selectionStart()
        sel_end = sel.cursor.selectionEnd()

        block_start, block_no_start = self.get_block(sel_start)
        block_end, block_no_end = self.get_block(sel_end)
        block_new, block_no_new = self.get_block(pos_new)

        if block_no_start == block_no_cur:
            if block_no_new <= block_no_end:
                block_start = block_new
            else:
                block_start = block_end
                block_end = block_new
        else:
            if block_no_new >= block_no_start:
                block_end = block_new
            else:
                block_end = block_start
                block_start = block_new

        start = block_start.position()
        end = block_end.position() + block_end.length() - 1
        sel.cursor.setPosition(start)
        sel.cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_selection", [sel])
        self.set_cursor_pos(pos_new)

    def set_block_selection_in_visual(self, motion_info: MotionInfo):
        """Update block-wise visual selection from ``motion_info``.

        Args:
            motion_info: Motion information computed by a helper.
        """
        editor = self.get_editor()
        sel = editor.get_extra_selections("vim_selection")[0]
        sel.cursor.setPosition(motion_info.sel_start)
        sel.cursor.setPosition(motion_info.sel_end, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_selection", [sel])
        self.set_cursor_pos(motion_info.sel_end - 1)

    def set_cursor_pos(self, pos):
        """Place the editor cursor at ``pos``.

        Args:
            pos: New absolute position for the cursor.
        """
        if pos is None:
            return
        editor = self.get_editor()
        pos = max(0, min(pos, editor.document().characterCount() - 1))
        cursor = editor.textCursor()
        cursor.setPosition(pos)
        self.set_cursor(cursor)
        self.draw_vim_cursor()

    def set_cursor_pos_without_end(self, pos):
        """Set the cursor avoiding the final block boundary.

        Args:
            pos: Desired cursor position.
        """
        if pos is None:
            return
        editor = self.get_editor()
        cursor = editor.textCursor()
        cursor.setPosition(pos)
        if cursor.atBlockEnd() and not cursor.atBlockStart():
            cursor.movePosition(QTextCursor.Left)
        self.set_cursor(cursor)
        self.draw_vim_cursor()

    def apply_motion_info_in_normal(self, motion_info: MotionInfo):
        """Apply motion info in normal mode."""
        self.set_cursor_pos(motion_info.cursor_pos)

    def apply_motion_info_in_yank(self, motion_info: MotionInfo):
        """Apply motion info after yank in normal mode."""
        cursor_pos = self.get_cursor().position()
        if motion_info.motion_type == MotionType.BlockWise:
            self.set_cursor_pos(motion_info.sel_start)
        else:
            if motion_info.cursor_pos is None:
                return
            pos = min([cursor_pos, motion_info.cursor_pos])
            self.set_cursor_pos(pos)

    def apply_motion_info_in_visual(self, motion_info: MotionInfo):
        """Apply motion info in visual mode."""
        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def apply_motion_info_in_vline(self, motion_info: MotionInfo):
        """Apply motion info in visual mode."""
        self.set_cursor_pos_in_vline(motion_info.cursor_pos)

    def set_extra_selections(
        self, key: str, sels: list[QTextEdit.ExtraSelection]
    ) -> None:
        """Set extra selections on the editor.

        Args:
            key: Name of the selection group.
            sels: List of selections to apply.
        """
        editor = self.get_editor()
        editor.set_extra_selections(key, sels)

    def highlight_yank(self, pos_start: int, pos_end: int) -> None:
        """Highlight yanked text.

        Args:
            pos_start: Starting position of the yanked text.
            pos_end: End position of the yanked text.
        """
        if self.hl_yank is False:
            return

        cursor = self.get_cursor()
        editor = self.get_editor()

        sel = QTextEdit.ExtraSelection()
        sel.format.setForeground(self.yank_fg_color)
        sel.format.setBackground(self.yank_bg_color)
        sel.cursor = cursor

        sel.cursor.setPosition(pos_start)
        sel.cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        self.set_extra_selections("hl_yank", [sel])

        def clear():
            try:
                editor.clear_extra_selections("hl_yank")
            except RuntimeError:
                pass

        QTimer.singleShot(self.hl_yank_dur, clear)



# -*- coding: utf-8 -*-
"""Implementation of EasyMotion overlays."""

# Standard Libraries
from itertools import product

# Third Party Libraries
from qtpy.QtCore import QCoreApplication, QEvent, QObject, Qt
from qtpy.QtGui import QColor, QPainter, QPen
from qtpy.QtWidgets import QPlainTextEdit, QWidget


class EasyMotionMarkerManager:
    """Manage overlay marker assignments."""

    def __init__(self):
        """Initialize empty marker sets."""
        self.position_list: list[int] = []
        self.name_list: list[str] = []

        self.marker_keys: list[str] = []
        self.init_marker_keys()
        self.motion_type = None

    def init_marker_keys(self):
        """Initialize markers."""
        keys1 = "hklyuiopnm,qwe"
        keys2 = "rtzxcvbasdgjf;"

        self.marker_keys = list(keys1)
        self.marker_keys += ["".join(pro) for pro in product(keys2, keys1)]

    def set_positions(self, position_list: list[int], motion_type):
        """Set positions."""
        self.name_list = []
        self.position_list = position_list
        self.name_list = self.marker_keys[: len(position_list)]
        self.motion_type = motion_type

    def handle_user_input(self, ch):
        """Process user input."""
        positions = []
        names = []
        for pos, name in zip(self.position_list, self.name_list):
            if name[0] != ch:
                continue
            positions.append(pos)
            if len(name) == 1:
                names.append(name)
                break
            else:
                names.append(name[1:])
        self.position_list = positions
        self.name_list = names


class EasyMotionPainter(QObject):
    """Draw overlay markers for EasyMotion in a text editor.

    Ref: github.com/serpheroth/QtCreator-EasyMotion

    """

    def __init__(self, editor: QPlainTextEdit = None):
        """Create painter.

        Args:
            editor: Widget receiving overlay rendering.
        """
        super().__init__()
        self.editor = editor
        self.positions = []
        self.names = []

    def eventFilter(self, viewport: QWidget, event: QEvent):
        if not viewport or event.type() != QEvent.Paint:
            return False
        # Handle the painter event last to prevent
        # the area painted by EasyMotion to be overridden
        viewport.removeEventFilter(self)
        QCoreApplication.sendEvent(viewport, event)
        viewport.installEventFilter(self)

        editor = self.editor
        tc = editor.textCursor()
        fm = editor.fontMetrics()

        pen_ch = QPen(QColor(220, 120, 120, 255))
        pen_border = QPen(QColor(0, 160, 100, 255), 1)

        with QPainter(viewport) as painter:
            painter.setCompositionMode(QPainter.CompositionMode_SourceIn)
            painter.setBrush(QColor(57, 34, 79, 255))
            painter.setFont(editor.font())
            ch_width = fm.width(" ")

            for pos_ch, marker_name in zip(self.positions, self.names):
                tc.setPosition(pos_ch)

                rect = editor.cursorRect(tc)
                rect.setWidth(ch_width * len(marker_name))

                if rect.intersects(viewport.rect()):
                    painter.setPen(pen_border)
                    painter.drawRoundedRect(rect, 40, 40, Qt.RelativeSize)
                    painter.setPen(pen_ch)
                    text_height = rect.bottom() - fm.descent()
                    painter.drawText(rect.left(), text_height, marker_name)

        return True



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
"""Tests for the plugin."""

# Third Party Libraries
import pytest
from qtpy.QtCore import QCoreApplication, QEvent, Qt
from qtpy.QtGui import QFocusEvent, QKeyEvent

# Project Libraries
from spyder_okvim.spyder.confpage import OkvimConfigPage


def test_conf_page(vim_bot):
    """Test conf_page.

    Call the methods that is difficult to make test case.
    """
    main, _, _, vim, _ = vim_bot
    vim.get_icon()
    vim.switch_to_plugin()
    conf_page = OkvimConfigPage(vim, main)
    conf_page.setup_page()

    old_leader_key = conf_page.leaderkey_viewer.textbox.text()
    new_event = QKeyEvent(QEvent.KeyPress, Qt.Key_Control, Qt.NoModifier)
    conf_page.leaderkey_edit.keyPressEvent(new_event)

    assert conf_page.leaderkey_viewer.textbox.text() == old_leader_key

    new_event = QKeyEvent(QEvent.KeyPress, Qt.Key_unknown, Qt.NoModifier)
    conf_page.leaderkey_edit.keyPressEvent(new_event)

    assert conf_page.leaderkey_viewer.textbox.text() == old_leader_key

    new_event = QKeyEvent(QEvent.KeyPress, Qt.Key_Meta, Qt.NoModifier)
    conf_page.leaderkey_edit.keyPressEvent(new_event)

    assert conf_page.leaderkey_viewer.textbox.text() == old_leader_key

    new_event = QKeyEvent(QEvent.KeyPress, Qt.Key_Enter, Qt.NoModifier)
    conf_page.leaderkey_edit.keyPressEvent(new_event)

    assert conf_page.leaderkey_viewer.textbox.text() == "Enter"


def test_apply_config(vim_bot):
    """Run apply_plugin_settings method."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("foo Foo foo Foo")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    # test refresh the color of search result.
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "/foo")
    qtbot.keyPress(cmd_line, Qt.Key_Return)

    vim.apply_plugin_settings("")


def test_ctrl_u_b(vim_bot):
    """Test ^u ^b."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "j")

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_U, Qt.ControlModifier)
    vim.vim_cmd.commandline.keyPressEvent(event)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == 0

    qtbot.keyClicks(cmd_line, "j")

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_B, Qt.ControlModifier)
    vim.vim_cmd.commandline.keyPressEvent(event)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == 0


def test_ctrl_d_f(vim_bot):
    """Test ^d ^f."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_D, Qt.ControlModifier)
    vim.vim_cmd.commandline.keyPressEvent(event)

    # assert cmd_line.text() == ""
    # assert editor.textCursor().position() == 2

    qtbot.keyClicks(cmd_line, "k")

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_F, Qt.ControlModifier)
    vim.vim_cmd.commandline.keyPressEvent(event)

    # assert cmd_line.text() == ""
    # assert editor.textCursor().position() == 2


def test_message(vim_bot):
    """Test message."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\nd\ne")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "y2j")
    assert vim.vim_cmd.msg_label.text() == "3 lines yanked"

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "p")
    assert vim.vim_cmd.msg_label.text() == "3 more lines"

    vim.vim_cmd.msg_label.clear()
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "vp")
    assert vim.vim_cmd.msg_label.text() == "4 more lines"

    vim.vim_cmd.msg_label.clear()
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "Vp")
    assert vim.vim_cmd.msg_label.text() == "2 more lines"

    qtbot.keyClicks(cmd_line, "i")
    cmd_line.focusInEvent(QFocusEvent(QEvent.FocusIn, Qt.OtherFocusReason))
    assert vim.vim_cmd.msg_label.text() == ""

    qtbot.keyClicks(cmd_line, "d2j")
    assert vim.vim_cmd.msg_label.text() == "3 fewer lines"

    editor.set_text("a\nb\nc\nd\ne")
    qtbot.keyClicks(cmd_line, "c2j")
    assert vim.vim_cmd.msg_label.text() == "2 fewer lines"

    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyPress(cmd_line, Qt.Key_Enter)
    assert vim.vim_cmd.msg_label.text() == ""

    vim.vim_cmd.msg_label.setText("a")
    qtbot.keyClicks(cmd_line, "/")
    qtbot.keyPress(cmd_line, Qt.Key_Enter)
    assert vim.vim_cmd.msg_label.text() == ""

    vim.vim_cmd.msg_label.setText("a")
    qtbot.keyClicks(cmd_line, "v/")
    qtbot.keyPress(cmd_line, Qt.Key_Enter)
    assert vim.vim_cmd.msg_label.text() == ""

    vim.vim_cmd.msg_label.setText("a")
    qtbot.keyClicks(cmd_line, "V/")
    qtbot.keyPress(cmd_line, Qt.Key_Enter)
    qtbot.keyPress(cmd_line, Qt.Key_Escape)
    assert vim.vim_cmd.msg_label.text() == ""

    vim.vim_cmd.msg_label.setText("a")
    editor.set_text("a\nb\nc\nd\ne")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    qtbot.keyClicks(cmd_line, "hx")
    qtbot.keyClicks(cmd_line, "u")
    assert vim.vim_cmd.msg_label.text() == "1 changes"

    vim.vim_cmd.msg_label.setText("a")
    qtbot.keyPress(cmd_line, Qt.Key_R, Qt.ControlModifier)
    assert vim.vim_cmd.msg_label.text() == "1 changes"

    qtbot.keyClicks(cmd_line, "d3j")
    qtbot.keyClicks(cmd_line, "u")
    assert vim.vim_cmd.msg_label.text() == "4 more lines"

    qtbot.keyPress(cmd_line, Qt.Key_R, Qt.ControlModifier)
    assert vim.vim_cmd.msg_label.text() == "4 fewer lines"

    qtbot.keyClicks(cmd_line, "p")
    qtbot.keyClicks(cmd_line, "u")
    assert vim.vim_cmd.msg_label.text() == "4 fewer lines"

    qtbot.keyPress(cmd_line, Qt.Key_R, Qt.ControlModifier)
    assert vim.vim_cmd.msg_label.text() == "4 more lines"

    # TODO: Spyder에서는 정상동작하지만 Pytest만 Fail
    # editor.set_text("aaaaaa")
    # vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    # vim.vim_cmd.vim_status.reset_for_test()
    # qtbot.keyClicks(cmd_line, "/a")
    # qtbot.keyPress(cmd_line, Qt.Key_Enter)
    # assert vim.vim_cmd.msg_label.text() == "/a"

    # vim.vim_cmd.msg_label.setText("a")
    # qtbot.keyClicks(cmd_line, "n")
    # assert vim.vim_cmd.msg_label.text() == "/a"

    # qtbot.keyClicks(cmd_line, "N")
    # assert vim.vim_cmd.msg_label.text() == "?a"


def test_ctrl_o_i(vim_bot):
    """Test jumplist navigation with ^O and ^I."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("alpha\nbravo\ncharlie\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "G")
    pos_bottom = editor.textCursor().position()

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_O, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == 0

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_I, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == pos_bottom


def test_jumplist_search(vim_bot):
    """Jump list records search command."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("alpha\nsearch\nsearch\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "/search")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    pos_search = editor.textCursor().position()

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_O, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == 0

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_I, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == pos_search


def test_jumplist_n_command(vim_bot):
    """Jump list records n command."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("foo foo foo foo")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "/foo")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    first = editor.textCursor().position()

    qtbot.keyClicks(cmd_line, "n")
    second = editor.textCursor().position()

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_O, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == first

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_I, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == second


def test_jumplist_mark_jump(vim_bot):
    """Jump list records mark jumps."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("alpha\nbravo\ncharlie\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    qtbot.keyClicks(cmd_line, "G")
    pos_end = editor.textCursor().position()

    qtbot.keyClicks(cmd_line, "'a")
    pos_mark = editor.textCursor().position()

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_O, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == pos_end

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_I, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == pos_mark


@pytest.mark.parametrize("line_num", [2, 3, 4])
def test_jumplist_colon_number(vim_bot, line_num):
    """Jump list records :number command for several lines."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(
        "alpha one two\n"
        "bravo charlie delta\n"
        "charlie echo foxtrot\n"
        "juliet kilo lima\n"
    )
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, f":{line_num}")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    pos_line = editor.textCursor().position()

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_O, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == 0

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_I, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert editor.textCursor().position() == pos_line


def test_jumplist_go_to_definition(vim_bot, monkeypatch):
    """Jump list records gd command."""
    _, stack, editor, vim, qtbot = vim_bot

    vs = vim.vim_cmd.vim_status
    stack.set_current_filename(stack.get_filenames()[0])
    vs.cursor.set_cursor_pos(0)
    vs.reset_for_test()

    other = next(p for p in stack.get_filenames() if p != stack.get_current_filename())

    # Third Party Libraries
    from spyder.plugins.editor.widgets.codeeditor import CodeEditor

    def fake_gd(self, cursor=None):
        stack.set_current_filename(other)
        vs.cursor.set_cursor_pos(1)

    monkeypatch.setattr(CodeEditor, "go_to_definition_from_cursor", fake_gd)

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "gd")
    qtbot.waitUntil(lambda: len(vs.jump_list.jumps) == 2, timeout=2000)
    QCoreApplication.processEvents()

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_O, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    assert stack.get_current_filename() != other
    assert editor.textCursor().position() == 0

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_I, Qt.ControlModifier)
    cmd_line.keyPressEvent(event)
    QCoreApplication.processEvents()
    qtbot.waitUntil(
        lambda: stack.get_current_filename() == other and vs.jump_list.index == 2,
        timeout=3000,
    )
    QCoreApplication.processEvents()
    assert stack.get_current_filename() == other
    # assert stack.get_current_editor().textCursor().position() == 1  # Fixme
    assert vs.jump_list.jumps[-1].pos == 1  # temprary fix for the above issue


def test_gd_no_change_pop(vim_bot, monkeypatch):
    """No jumplist entry kept when gd doesn't move."""
    _, stack, _, vim, qtbot = vim_bot
    vs = vim.vim_cmd.vim_status
    stack.set_current_filename(stack.get_filenames()[0])
    vs.cursor.set_cursor_pos(0)
    vs.reset_for_test()

    # Third Party Libraries
    from spyder.plugins.editor.widgets.codeeditor import CodeEditor

    def fake_gd(self, cursor=None):
        pass

    monkeypatch.setattr(CodeEditor, "go_to_definition_from_cursor", fake_gd)

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "gd")
    qtbot.wait(2500)

    assert vs.jump_list.jumps == []



"""Conftest."""

# Standard Libraries
import os
import os.path as osp
from unittest.mock import Mock

os.environ.setdefault("QT_QPA_PLATFORM", "offscreen")

# Third Party Libraries
import pytest
from pytestqt.plugin import QtBot
from qtpy.QtCore import QCoreApplication
from qtpy.QtGui import QFont
from qtpy.QtWidgets import QVBoxLayout, QWidget
from spyder.api.plugins import Plugins
from spyder.config.manager import CONF
from spyder.plugins.editor.widgets.editorstack import EditorStack

# Project Libraries
from spyder_okvim.spyder.config import CONF_DEFAULTS, CONF_SECTION
from spyder_okvim.spyder.plugin import OkVim

LOCATION = osp.realpath(osp.join(os.getcwd(), osp.dirname(__file__)))


class VimTesting(OkVim):
    CONF_FILE = False
    CONF_SECTION = CONF_SECTION


class EditorMock(QWidget):
    """Editor plugin mock."""

    def __init__(self, editor_stack):
        """Editor Mock constructor."""
        QWidget.__init__(self, None)
        self.editor_stack = editor_stack
        self.editorsplitter = self.editor_stack
        self.open_action = Mock()
        self.new_action = Mock()
        self.new = Mock()
        self.save_action = Mock()
        self.save = Mock()
        self.close_action = Mock()
        self.close_file = Mock()

        layout = QVBoxLayout()
        layout.addWidget(self.editor_stack)
        self.setLayout(layout)

    def get_widget(self):
        """Return self, mimicking Spyder 6 API."""
        return self

    def get_current_editorstack(self):
        """Return EditorStack instance."""
        return self.editor_stack


class StatusBarMock(QWidget):
    """Stausbar mock."""

    def __init__(self):
        QWidget.__init__(self, None)
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

    def add_status_widget(self, widget):
        self.layout.addWidget(widget)


class MainMock(QWidget):
    """Spyder MainWindow mock."""

    def __init__(self, editor_stack, qtbot_module):
        """Main Window Mock constructor."""
        QWidget.__init__(self, None)
        self.main = QWidget()
        self.plugin_focus_changed = Mock()
        self.editor = EditorMock(editor_stack)
        self.status_bar = StatusBarMock()
        self.application = Mock()
        self.application.open_file_in_plugin = Mock()
        self.projects = Mock()
        self.load_edit = Mock()
        self.projects.get_active_project_path = lambda: None
        layout = QVBoxLayout()
        layout.addWidget(self.editor)
        layout.addWidget(self.status_bar)
        self.setLayout(layout)

        for name, dict_info in CONF_DEFAULTS:
            if name != CONF_SECTION:
                continue
            for key, val in dict_info.items():
                CONF.set(name, key, val)

        # Disable highlight yank delay to avoid timer side effects
        CONF.set(CONF_SECTION, "highlight_yank_duration", 0)

        self.add_dockwidget = Mock()

        qtbot_module.add_widget(self.main)
        qtbot_module.add_widget(self.editor)
        qtbot_module.add_widget(self.status_bar)

    def get_plugin(self, plugin, error=True):
        if plugin == Plugins.StatusBar:
            return self.status_bar
        if plugin == Plugins.Application:
            return self.application
        return self.status_bar


@pytest.fixture(scope="session")
def qtbot_module(qapp, request):
    """Module fixture for qtbot."""
    result = QtBot(request)
    return result


@pytest.fixture(scope="session", autouse=True)
def vim_bot(qtbot_module):
    """Editorstack pytest fixture."""
    text = (
        "   123\n" "line 1\n" "line 2\n" "line 3\n" "line 4"
    )  # a newline is added at end
    editor_stack = EditorStack(None, [])

    # Fix the area of the selection
    font = QFont("Courier New")
    font.setPointSize(10)
    editor_stack.set_default_font(font)
    editor_stack.setMinimumWidth(400)
    editor_stack.setMinimumHeight(400)

    editor_stack.set_find_widget(Mock())
    if hasattr(editor_stack, "set_io_actions"):
        editor_stack.set_io_actions(Mock(), Mock(), Mock(), Mock())
    finfo0 = editor_stack.new(osp.join(LOCATION, "foo.py"), "utf-8", text)
    finfo1 = editor_stack.new(osp.join(LOCATION, "foo1.py"), "utf-8", text)
    finfo2 = editor_stack.new(osp.join(LOCATION, "foo2.py"), "utf-8", text)
    finfo3 = editor_stack.new(osp.join(LOCATION, "foo3.py"), "utf-8", text)
    main = MainMock(editor_stack, qtbot_module)

    vim = VimTesting(main, None)
    vim.on_initialize()

    qtbot_module.addWidget(editor_stack)
    qtbot_module.addWidget(finfo0.editor)
    qtbot_module.addWidget(finfo1.editor)
    qtbot_module.addWidget(finfo2.editor)
    qtbot_module.addWidget(finfo3.editor)
    qtbot_module.addWidget(main)

    # Show GUI
    main.show()
    yield main, editor_stack, finfo0.editor, vim, qtbot_module

    finfo0.editor.close()
    finfo1.editor.close()
    finfo2.editor.close()
    finfo3.editor.close()
    finfo0.deleteLater()
    finfo1.deleteLater()
    finfo2.deleteLater()
    finfo3.deleteLater()
    editor_stack.deleteLater()
    # Close plugin widgets to avoid segmentation faults under Qt
    vim.on_close()
    main.close()
    main.deleteLater()
    # Ensure deferred deletions are processed to avoid segmentation faults
    QCoreApplication.processEvents()


@pytest.fixture(autouse=True)
def process_events_after_test():
    """Process pending Qt events after each test."""
    yield
    QCoreApplication.processEvents()



# -*- coding: utf-8 -*-
"""Executor for ":" command-line input."""

# Third Party Libraries
from qtpy.QtWidgets import QDialog

# Project Libraries
from spyder_okvim.executor.executor_base import ExecutorSubBase
from spyder_okvim.utils.jump_dialog import JumpListDialog
from spyder_okvim.utils.mark_dialog import MarkListDialog
from spyder_okvim.vim import VimState


class ExecutorColon(ExecutorSubBase):
    """Interpreter for ``:`` commands entered in the command line."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.editor_widget = vim_status.editor_widget

    def __call__(self, txt: str) -> bool:
        """Parse ``txt`` and execute the corresponding ex command.

        Args:
            txt: Command string typed by the user, including the trailing ``\r``.

        Returns:
            if return is True, Clear command line.

        """
        if txt[-1] == "\b":
            cmd_line = self.vim_status.cmd_line
            if len(txt) <= 2:
                self.vim_status.sub_mode = None
                return True
            else:
                cmd_line.setText(txt[:-2])
        if txt[-1] != "\r":
            return False

        if txt[0] != ":" or len(txt) <= 2:
            self.vim_status.sub_mode = None
            return True

        # ':' is saved to input cmd_info at ExecutorBase.
        # So we need only txt[1:].
        self.update_input_cmd_info(None, None, txt[1:])

        txt = txt[1:-1]  # remove :, \r
        cmd = txt.split(None, 1)
        args = cmd[1] if len(cmd) > 1 else ""
        cmd = cmd[0]

        if cmd.isdigit():
            self.goto_line(int(cmd))
        else:
            for symbol, text in self.SYMBOLS_REPLACEMENT.items():
                cmd = cmd.replace(symbol, text)

            try:
                method = self.__getattribute__(cmd)
            except AttributeError:
                print("unknown command", cmd)
            else:
                method(args)

        self.vim_status.sub_mode = None

        return True

    def w(self, arg=""):
        """Save current file."""
        self.save_current_file()
        self.vim_status.cursor.draw_vim_cursor()

    def q(self, arg=""):
        """Close current file."""
        self.close_current_file()
        self.vim_status.set_focus_to_vim()

    def qexclamation(self, arg=""):
        """Close current file without saving."""
        # TODO :
        self.close_current_file()
        self.vim_status.set_focus_to_vim()

    def wq(self, arg=""):
        """Save and close current file."""
        self.w(arg)
        self.q()

    def n(self, args=""):
        """Create new file."""
        self.create_new_file()
        self.vim_status.set_focus_to_vim()

    def marks(self, arg=""):
        """Show bookmarks and jump to the selected one."""
        vs = self.vim_status
        marks = vs.bookmark_manager.list_bookmarks()

        dlg = MarkListDialog(marks, vs.main)
        result = dlg.exec_()
        # When running tests ``exec_`` is patched to return ``None`` so the
        # selected mark must be retrieved regardless of the returned value.
        # Only reject the action explicitly when ``QDialog.Rejected`` (0) is
        # returned.
        mark = (
            dlg.get_selected_mark()
            if result is None or result == QDialog.Accepted
            else None
        )
        if mark:
            vs.push_jump()
            vs.jump_to_bookmark(mark)
            vs.push_jump()

        vs.set_focus_to_vim()

    def jumps(self, arg=""):
        """Display the jumplist in a popup dialog."""
        vs = self.vim_status

        dlg = JumpListDialog(vs, vs.main)
        dlg.exec_()

        vs.set_focus_to_vim()

    def goto_line(self, num):
        """Move cursor according to :number command."""
        vs = self.vim_status
        vs.push_jump()
        if vs.vim_state == VimState.NORMAL:
            motion_info = self.helper_motion.G(num, True)
            vs.cursor.apply_motion_info_in_normal(motion_info)
        elif vs.vim_state in (VimState.VISUAL, VimState.VLINE):
            motion_info = self.helper_motion.j(num)
            if vs.vim_state == VimState.VISUAL:
                vs.cursor.apply_motion_info_in_visual(motion_info)
            else:
                vs.cursor.apply_motion_info_in_vline(motion_info)
            vs.to_normal()
        vs.push_jump()



# -*- coding: utf-8 -*-
"""Vertical line selection helper."""

# Standard Libraries
import re

# Third Party Libraries
from spyder.config.manager import CONF

# Project Libraries
from spyder_okvim.executor.executor_base import (
    FUNC_INFO,
    RETURN_EXECUTOR_METHOD_INFO,
    ExecutorBase,
)
from spyder_okvim.executor.decorators import submode
from spyder_okvim.executor.executor_colon import ExecutorColon
from spyder_okvim.executor.executor_easymotion import ExecutorEasymotion
from spyder_okvim.executor.executor_sub import (
    ExecutorSearch,
    ExecutorSubCmd_alnum,
    ExecutorSubCmd_closesquarebracket,
    ExecutorSubCmd_closebrace,
    ExecutorSubCmd_f_t,
    ExecutorSubCmd_g,
    ExecutorSubCmd_opensquarebracket,
    ExecutorSubCmd_openbrace,
    ExecutorSubCmd_r,
    ExecutorSubCmd_register,
    ExecutorSubCmdSneak,
)
from spyder_okvim.executor.mixins import MovementMixin, SelectionMixin
from spyder_okvim.spyder.config import CONF_SECTION


class ExecutorVlineCmd(SelectionMixin, MovementMixin, ExecutorBase):
    """Executor for vline."""

    def __init__(self, vim_status):
        super().__init__(vim_status)

        # MovementMixin hooks
        self.move_cursor = vim_status.cursor.set_cursor_pos_in_vline
        self.move_cursor_no_end = vim_status.cursor.set_cursor_pos_in_vline

        cmds = "uUovhydcsSxHjJklLMwWbBepP^$gG~:%fFtTnN/;,\"`'mr<>{} \b\r[]*#"
        cmds = "".join(re.escape(c) for c in cmds)
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))
        self.set_cursor_pos = vim_status.cursor.set_cursor_pos
        self.set_cursor_pos_in_vline = vim_status.cursor.set_cursor_pos_in_vline
        self.apply_motion_info_in_vline = (
            self.vim_status.cursor.apply_motion_info_in_vline
        )
        self.executor_sub_g = ExecutorSubCmd_g(vim_status)
        self.executor_sub_f_t = ExecutorSubCmd_f_t(vim_status)
        self.executor_sub_r = ExecutorSubCmd_r(vim_status)
        self.executor_sub_register = ExecutorSubCmd_register(vim_status)
        self.executor_sub_alnum = ExecutorSubCmd_alnum(vim_status)
        self.executor_sub_search = ExecutorSearch(vim_status)
        self.executor_sub_easymotion = ExecutorEasymotion(vim_status)
        self.executor_sub_sneak = ExecutorSubCmdSneak(vim_status)
        self.executor_colon = ExecutorColon(vim_status)
        self.executor_sub_opensquarebracekt = ExecutorSubCmd_opensquarebracket(
            vim_status
        )
        self.executor_sub_closesquarebracekt = (
            ExecutorSubCmd_closesquarebracket(vim_status)
        )
        self.executor_sub_openbrace = ExecutorSubCmd_openbrace(vim_status)
        self.executor_sub_closebrace = ExecutorSubCmd_closebrace(vim_status)

        # SelectionMixin hooks
        self.apply_motion_info_in_sel = self.apply_motion_info_in_vline
        self.set_cursor_pos_in_sel = self.set_cursor_pos_in_vline
        self.paste_in_sel = self.helper_action.paste_in_vline

    def v(self, num=1, num_str=""):
        """Start Visual mode per character."""
        self.vim_status.to_visual_char()

    def J(self, num=1, num_str=""):
        """Join lines, with a minimum of two lines."""
        block_no_start = self.get_block_no_start_in_selection()
        block_no_end = self.get_block_no_end_in_selection()

        if block_no_start is None:
            self.vim_status.to_normal()
            return

        if block_no_start == block_no_end:
            block_no_end += 1

        cursor_pos_start = self.get_pos_start_in_selection()
        self.helper_action.join_lines(cursor_pos_start, block_no_start, block_no_end)
        self.vim_status.to_normal()

    def g(self, num=1, num_str=""):
        """Start g submode."""
        executor_sub = self.executor_sub_g

        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_vline, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def G(self, num=1, num_str=""):
        """Move to the line."""
        motion_info = self.helper_motion.G(num=num, num_str=num_str)

        self.set_cursor_pos_in_vline(motion_info.cursor_pos)

    def tilde(self, num=1, num_str=""):
        """Switch case of the character under the cursor.

        Move the cursor to the start of selection.
        """
        self.helper_action.handle_case(None, "swap")

    def f(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_vline, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def F(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_vline, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def t(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_vline, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def T(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_vline, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_vline, True)])
    def opensquarebracket(self, num=1, num_str=""):
        """Start [ submode."""
        return self.executor_sub_opensquarebracekt

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_vline, True)])
    def closesquarebracket(self, num=1, num_str=""):
        """Start ] submode."""
        return self.executor_sub_closesquarebracekt

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_vline, True)])
    def openbrace(self, num=1, num_str=""):
        """Start { submode."""
        return self.executor_sub_openbrace

    @submode(lambda self: [FUNC_INFO(self.apply_motion_info_in_vline, True)])
    def closebrace(self, num=1, num_str=""):
        """Start } submode."""
        return self.executor_sub_closebrace

    def d(self, num=1, num_str=""):
        """Delete text."""
        editor = self.get_editor()
        block_no_start = self.get_block_no_start_in_selection()

        self.helper_action.yank(None)
        self.helper_action.delete(None, is_insert=False)
        self.vim_status.to_normal()

        # calc cursor position
        n_block = editor.blockCount()
        block_no = min([block_no_start, n_block - 1])
        block = editor.document().findBlockByNumber(block_no)
        cursor_pos = block.position()
        txt = block.text()
        cursor_pos += len(txt) - len(txt.lstrip())

        self.set_cursor_pos(cursor_pos)

    def S(self, num=1, num_str=""):
        """Delete text and start insert."""
        use_sneak = CONF.get(CONF_SECTION, "use_sneak")
        if use_sneak:
            executor_sub = self.executor_sub_sneak

            self.set_parent_info_to_submode(executor_sub, num, num_str)

            executor_sub.set_func_list_deferred(
                [
                    FUNC_INFO(self.apply_motion_info_in_vline, True),
                    FUNC_INFO(
                        self.helper_motion.display_another_group_after_rsneak, False
                    ),
                ]
            )

            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)



# -*- coding: utf-8 -*-
"""Tests for the executor_vline."""

# Third Party Libraries
import pytest
from qtpy.QtCore import Qt
from spyder.config.manager import CONF

# Project Libraries
from spyder_okvim.spyder.config import CONF_SECTION
from spyder_okvim.vim import VimState


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("import numpy as np", ["V"], 0, [0, 18]),
        ("import numpy as np", ["V", "l", "h"], 0, [0, 18]),
        ("import numpy as np", ["V", "5l"], 5, [0, 18]),
        ("import numpy as np", ["5l", "V"], 5, [0, 18]),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["V", "j"],
            19,
            [0, 50],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["2l", "V", "2j", "5l"],
            58,
            [0, 75],
        ),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc
""",
            ["2l", "V", "3j", "k", "j"],
            76,
            [0, 76],
        ),
        (
            """
import matplotlib.pyplot as plt
import scipy.scipy as sc
""",
            ["5j", "V", "5k"],
            0,
            [0, 58],
        ),
    ],
)
def test_V_cmd(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test V command."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("import numpy as np", ["V", "0"], 0, [0, 18]),
        ("import numpy as np", ["5l", "V", "0"], 0, [0, 18]),
        (
            """import numpy as np
import matplotlib.pyplot as plt
import scipy.scipy as sc""",
            ["V", "j", "5l", "0"],
            19,
            [0, 50],
        ),
    ],
)
def test_zero_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test 0 command in v-line."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("   import numpy as np", ["V", "^"], 3, [0, 21]),
        ("   import numpy as np", ["10l", "V", "^"], 3, [0, 21]),
    ],
)
def test_caret_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test ^ command in v-line."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [("import numpy as np", ["V", "$"], 18, [0, 18])],
)
def test_dollar_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test $ command in v-line."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34\n", ["V", "w"], 3, [0, 5]),
        ("01 34\n", ["V", "w", "o"], 0, [0, 5]),
        ("01 34\n", ["V", "w", "o", "o"], 4, [0, 5]),
        ("\n", ["j", "V"], 1, [1, 1]),
        ("\n", ["j", "V", "o"], 1, [1, 1]),
        ("01 34\n6\n", ["V", "j"], 6, [0, 7]),
        ("01 34\n6\n", ["V", "j", "o"], 0, [0, 7]),
        ("01 34\n6\n", ["j", "V", "k", "o"], 6, [0, 7]),
        ("01 34\n6\n8\n", ["j", "V", "j", "2k", "o"], 6, [0, 7]),
    ],
)
def test_o_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test o command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("", ["V", "J"], "", 0),
        ("\n\n", ["3j", "V", "J"], "\n\n", 2),
        ("0\n23", ["j", "l", "V", "k", "J"], "0 23", 1),
        ("0\n2\n4\n6\n8\n", ["V", "2j", "J"], "0 2 4\n6\n8\n", 3),
        ("0\n2\n4\n6\n8\n", ["V", "2j", "J", "."], "0 2 4 6 8\n", 7),
    ],
)
def test_J_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test J command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["V", "w"], 3, [0, 5]),
    ],
)
def test_w_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test w command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["V", "W"], 0, [0, 0]),
        ("029.d98@jl 34", ["V", "W"], 11, [0, 13]),
        ("029.d98@jl 34", ["V", "2W"], 13, [0, 13]),
        ("029.d98@jl 34\na", ["V", "2W"], 14, [0, 15]),
        ("029.d98@jl 34\n  a", ["V", "2W"], 16, [0, 17]),
    ],
)
def test_W_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test W command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["$", "V", "b"], 3, [0, 5]),
    ],
)
def test_b_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test b command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01.34", ["$", "V", "B"], 0, [0, 5]),
    ],
)
def test_B_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test B command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34\n67 90", ["V", "e"], 1, [0, 5]),
        ("01 34\n67 90", ["V", "3e"], 7, [0, 11]),
    ],
)
def test_e_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test e command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("0\n2\n4\n", ["V", "2G"], 2, [0, 3]),
        ("0\n     \n8\n", ["V", "2G"], 6, [0, 7]),
        ("0\n2\n4\n", ["V", "G"], 6, [0, 6]),
        ("0\n2\n4\n     a", ["V", "G"], 11, [0, 12]),
    ],
)
def test_G_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test G command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("0\n2\n4\n", ["V", "2gg"], 2, [0, 3]),
        ("0\n     \n8\n", ["V", "2gg"], 6, [0, 7]),
        ("    0\n2\n4\n", ["4j", "V", "gg"], 4, [0, 10]),
    ],
)
def test_gg_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test gg command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos",
    [
        ("0\n2\n4\n", ["V", ":", "2", Qt.Key_Return], 4),
        ("0\n     \n8\n", ["V", ":", "2", Qt.Key_Return], 8),
        ("0\n\n2\n4\n", ["V", ":", "3", Qt.Key_Return], 5),
        (
            "aaa\nbbb\nccc\nddd\neee\n",
            ["j", "V", ":", "3", Qt.Key_Return],
            16,
        ),
        (
            "aaa\nbbb\nccc\nddd\neee\n",
            ["j", "l", "V", ":", "3", Qt.Key_Return],
            17,
        ),
    ],
)
def test_colon_num_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos):
    """Test :num command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert not editor.get_extra_selections("vim_selection")


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("abcde", ["V", "~"], "ABCDE", 0),
        ("abcde\na", ["l", "V", "$", "~"], "ABCDE\na", 0),
        ("abcde\na", ["l", "V", "$", "~", "j", "."], "ABCDE\nA", 6),
        ("", ["V", "~"], "", 0),
    ],
)
def test_tilde_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test ~ command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["V", "%"], 0, [0, 0]),
        ("\n", ["j", "V", "%"], 1, [1, 1]),
        (" ()", ["V", "%"], 2, [0, 3]),
        (" ()", ["V", "%", "%"], 1, [0, 3]),
    ],
)
def test_percent_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test % command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["V", "f", "r"], 0, [0, 0]),
        ("\n", ["j", "V", "f", "r"], 1, [1, 1]),
        (" rr", ["V", "f", "r"], 1, [0, 3]),
        (" rr", ["V", "f", "r", ";"], 2, [0, 3]),
        (" rr", ["V", "f", "r", ";", ","], 1, [0, 3]),
    ],
)
def test_f_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test f command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["V", "F", "r"], 0, [0, 0]),
        ("\n", ["j", "V", "F", "r"], 1, [1, 1]),
        (" rr ", ["V", "$", "F", "r"], 2, [0, 4]),
        (" rr ", ["V", "$", "F", "r", ";"], 1, [0, 4]),
        (" rr ", ["V", "$", "F", "r", ";", ","], 2, [0, 4]),
    ],
)
def test_F_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test F command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["V", "s", "aa"], 0, [0, 0]),
        ("", ["V", ";"], 0, [0, 0]),
        ("", ["V", ","], 0, [0, 0]),
        ("\n", ["V", "j", "s", "rr"], 1, [0, 1]),
        ("d\ndhr Dhr dhr", ["V", "sdh"], 2, [0, 13]),
        ("d\ndhr Dhr dhr", ["V", "sdh", ";"], 10, [0, 13]),
        ("d\ndhr Dhr dhr", ["V", "sdh", ";,"], 2, [0, 13]),
        ("d\ndhr Dhr dhr", ["V", "j$" "Sdh"], 10, [0, 13]),
        ("d\ndhr Dhr dhr", ["V", "j$" "Sdh", ";"], 2, [0, 13]),
        ("d\ndhr Dhr dhr", ["V", "j$" "Sdh", ";,"], 10, [0, 13]),
    ],
)
def test_sneak_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test sneak command in visual."""
    CONF.set(CONF_SECTION, "use_sneak", True)
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["V", "t", "r"], 0, [0, 0]),
        ("\n", ["j", "V", "t", "r"], 1, [1, 1]),
        ("  rr", ["V", "t", "r"], 1, [0, 4]),
        ("  rr", ["V", "t", "r", ";"], 2, [0, 4]),
        ("  rrrr", ["V", "t", "r", "4;"], 4, [0, 6]),
        ("  rrrr", ["V", "t", "r", "4;", ","], 3, [0, 6]),
    ],
)
def test_t_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test t command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("", ["V", "T", "r"], 0, [0, 0]),
        ("r\n", ["j", "V", "T", "r"], 2, [2, 2]),
        ("  rr  ", ["V", "$", "T", "r"], 4, [0, 6]),
        ("  rr  ", ["V", "$", "T", "r", ";"], 3, [0, 6]),
        ("  rrrr", ["V", "$", "T", "r", "4;"], 3, [0, 6]),
        ("  rrrr", ["V", "$", "T", "r", "4;", ","], 4, [0, 6]),
    ],
)
def test_T_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test T command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("", ["V", "r", "r"], "", 0),
        ("1\n", ["j", "V", "r", "r"], "1\n", 2),
        ("\n\na", ["j", "V", "r", "r"], "\n\na", 1),
        ("a", ["V", "r", "r"], "r", 0),
        (" a\nbc\n", ["l", "V", "j", "r", "r"], "rr\nrr\n", 0),
        (" a\nbc\nkk", ["l", "V", "j", "r", "r"], "rr\nrr\nkk", 0),
        (" a\nbc\nkk", ["l", "V", "j", "r", "r", "j", "."], "rr\nrr\nrr", 3),
    ],
)
def test_r_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test r command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")

    assert sel == []
    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.vim_state == VimState.NORMAL
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("ABCDE", ["V", "u"], "abcde", 0),
        ("ABCDE\nA", ["l", "V", "$", "u"], "abcde\nA", 0),
        ("ABCDE\nA", ["l", "V", "$", "u", "j", "."], "abcde\na", 6),
        ("", ["V", "u"], "", 0),
    ],
)
def test_u_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test u command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("ABCDE", ["V", "g", "u"], "abcde", 0),
        ("ABCDE\nA", ["l", "V", "$", "g", "u"], "abcde\nA", 0),
        ("ABCDE\nA", ["l", "V", "$", "g", "u", "j", "."], "abcde\na", 6),
        ("", ["V", "g", "u"], "", 0),
    ],
)
def test_gu_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test gu command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("abcde", ["V", "U"], "ABCDE", 0),
        ("abcde\na", ["l", "V", "$", "U"], "ABCDE\na", 0),
        ("abcde\na", ["l", "V", "$", "U", "j", "."], "ABCDE\nA", 6),
        ("", ["V", "U"], "", 0),
    ],
)
def test_U_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test U command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("abcde", ["V", "g", "U"], "ABCDE", 0),
        ("abcde\na", ["l", "V", "$", "g", "U"], "ABCDE\na", 0),
        ("abcde\na", ["l", "V", "$", "g", "U", "j", "."], "ABCDE\nA", 6),
        ("", ["V", "g", "U"], "", 0),
    ],
)
def test_gU_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test gU command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("abCde", ["V", "g", "~"], "ABcDE", 0),
        ("abCde\na", ["l", "V", "$", "g", "~"], "ABcDE\na", 0),
        ("abCde\na", ["l", "V", "$", "g", "~", "j", "."], "ABcDE\nA", 6),
        ("", ["V", "g", "~"], "", 0),
    ],
)
def test_gtilde_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test g~ command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("", ["V", ">"], "", 0),
        ("abcde", ["2l", "V", ">"], "    abcde", 4),
        (" abcde\na", ["V", ">"], "     abcde\na", 5),
        ("a\n\na", ["V", "2j", ">"], "    a\n\n    a", 4),
    ],
)
def test_greater_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test > command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, text_expected, cursor_pos",
    [
        ("", ["V", "<"], "", 0),
        ("    abcde", ["2l", "V", "<"], "abcde", 0),
        ("     abcde\na", ["V", "<"], " abcde\na", 1),
        ("    a\n\n    a", ["V", "2j", "<"], "a\n\na", 0),
    ],
)
def test_less_cmd_in_vline(vim_bot, text, cmd_list, text_expected, cursor_pos):
    """Test < command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, register_name, text_yanked",
    [
        ("a", ["V", "y"], 0, '"', "a\n"),
        ("abcd", ["V", '"', "0", "y"], 0, "0", "abcd\n"),
        ("abcd\ne", ["V", "j", '"', "a", "y"], 0, "a", "abcd\ne\n"),
    ],
)
def test_y_cmd_in_vline(
    vim_bot, text, cmd_list, cursor_pos, register_name, text_yanked
):
    """Test y command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[register_name]
    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert reg.content == text_yanked
    assert reg.type == VimState.VLINE
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None
    if register_name == '"':
        reg0 = vim.vim_cmd.vim_status.register_dict["0"]
        assert reg0.content == text_yanked


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected",
    [
        ("ak", ["V", "p"], 0, ""),
        ("ak", ["v", "l", "y", "V", "p"], 0, "ak"),
        ("ak", ["v", "l", "y", "V", "P"], 0, "ak"),
        ("ak", ["v", "l", "y", "V", "2p"], 0, "ak\nak"),
        ("ab\n\ncd\n", ["v", "j", "y", "2j", "V", "2p"], 4, "ab\n\nab\n\n\nab\n\n\n"),
        (
            "ab\ncd\nef\ngh\n",
            ["V", "j", "y", "2j", "V", "p"],
            6,
            "ab\ncd\nab\ncd\ngh\n",
        ),
        (
            "ab\ncd\nef\ngh\n",
            ["V", "j", "y", "2j", "V", "2p"],
            6,
            "ab\ncd\nab\ncd\nab\ncd\ngh\n",
        ),
    ],
)
def test_p_P_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, text_expected):
    """Test p command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert editor.toPlainText() == text_expected
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        ("ab", ["V", "d"], 0, "", '"', "ab\n"),
        (" ab\n cd\n ef", ["j", "V", "d"], 5, " ab\n ef", '"', " cd\n"),
        (" ab\n cd\n", ["V", "G", "d"], 0, "", '"', " ab\n cd\n\n"),
        (" ab\n cd\n ef", ["2j", "V", "d"], 5, " ab\n cd", '"', " ef\n"),
        (" ab\n cd\n ef", ["2j", "V", "k", "d"], 1, " ab", '"', " cd\n ef\n"),
        (" ab\n cd", ["$", "V", "d"], 1, " cd", '"', " ab\n"),
    ],
)
def test_d_cmd_in_vline(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test d command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert reg.content == text_yanked
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        ("ab", ["V", "x"], 0, "", '"', "ab\n"),
        (" ab\n cd\n ef", ["j", "V", "x"], 5, " ab\n ef", '"', " cd\n"),
        (" ab\n cd\n", ["V", "G", "x"], 0, "", '"', " ab\n cd\n\n"),
        (" ab\n cd\n ef", ["2j", "V", "x"], 5, " ab\n cd", '"', " ef\n"),
        (" ab\n cd\n ef", ["2j", "V", "k", "x"], 1, " ab", '"', " cd\n ef\n"),
        (" ab\n cd", ["$", "V", "x"], 1, " cd", '"', " ab\n"),
    ],
)
def test_x_cmd_in_vline(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test x command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert reg.content == text_yanked
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        ("ab", ["V", "c"], 0, "", '"', "ab\n"),
        (" ab\n cd\n ef", ["j", "V", "c"], 4, " ab\n\n ef", '"', " cd\n"),
        (" ab\n cd\n", ["V", "G", "c"], 0, "", '"', " ab\n cd\n\n"),
        (" ab\n cd\n ef", ["2j", "V", "c"], 8, " ab\n cd\n", '"', " ef\n"),
        (" ab\n cd\n ef", ["2j", "V", "k", "c"], 4, " ab\n", '"', " cd\n ef\n"),
        (" ab\n cd", ["$", "V", "c"], 0, "\n cd", '"', " ab\n"),
    ],
)
def test_c_cmd_in_vline(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test c command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert reg.content == text_yanked
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked",
    [
        ("ab", ["V", "s"], 0, "", '"', "ab\n"),
        (" ab\n cd\n ef", ["j", "V", "s"], 4, " ab\n\n ef", '"', " cd\n"),
        (" ab\n cd\n", ["V", "G", "s"], 0, "", '"', " ab\n cd\n\n"),
        (" ab\n cd\n ef", ["2j", "V", "s"], 8, " ab\n cd\n", '"', " ef\n"),
        (" ab\n cd\n ef", ["2j", "V", "k", "s"], 4, " ab\n", '"', " cd\n ef\n"),
        (" ab\n cd", ["$", "V", "s"], 0, "\n cd", '"', " ab\n"),
    ],
)
def test_s_cmd_in_vline(
    vim_bot, text, cmd_list, cursor_pos, text_expected, reg_name, text_yanked
):
    """Test s command in vline."""
    CONF.set(CONF_SECTION, "use_sneak", False)
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    reg = vim.vim_cmd.vim_status.register_dict[reg_name]
    assert cmd_line.text() == ""
    assert editor.toPlainText() == text_expected
    assert reg.content == text_yanked
    assert editor.textCursor().position() == cursor_pos
    assert vim.vim_cmd.vim_status.get_pos_start_in_selection() is None


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("a", ["V", "/", "b", "\r"], 0, [0, 1]),
        ("a", ["V", "/", "b", "\r", "n"], 0, [0, 1]),
        (
            " dhr\n  dhrwodn\n\ndhrwodn\n   dhrwodn",
            ["V", "/", "d", "h", "r", Qt.Key_Enter],
            1,
            [0, 4],
        ),
        (
            " dhr\n  dhrwodn\n\ndhrwodn\n   dhrwodn",
            ["V", "/", "d", "h", "r", Qt.Key_Enter, "n"],
            7,
            [0, 14],
        ),
        (
            " dhr\n  dhrwodn\n\ndhrwodn\n   dhrwodn",
            ["V", "/", "d", "h", "r", Qt.Key_Return, "n", "N"],
            1,
            [0, 4],
        ),
    ],
)
def test_search_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test / command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos == sel_pos_


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("a", ["V", "*"], 0, [0, 1]),
        ("a", ["V", "#"], 0, [0, 1]),
        ("dhr\n  Dhr\n\n_dhr\n   dhr", ["V", "*"], 19, [0, 22]),
        ("dhr\n  dhr\n\n_dhr\n   dhr", ["j2l", "V", "#"], 0, [0, 9]),
    ],
)
def test_asterisk_sharp_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test *,# command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos == sel_pos_


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["V", Qt.Key_Space], 1, [0, 5]),
    ],
)
def test_space_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test space command in vline."""
    _, _, editor, vim, qtbot = vim_bot

    CONF.set(CONF_SECTION, "leader_key", "F1")
    vim.apply_plugin_settings("")

    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34", ["2l", "V", Qt.Key_Backspace], 1, [0, 5]),
    ],
)
def test_backspace_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test backspace command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


@pytest.mark.parametrize(
    "text, cmd_list, cursor_pos, sel_pos",
    [
        ("01 34\n   kj", ["V", Qt.Key_Enter], 9, [0, 11]),
    ],
)
def test_enter_cmd_in_vline(vim_bot, text, cmd_list, cursor_pos, sel_pos):
    """Test enter command in vline."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    sel = editor.get_extra_selections("vim_selection")[0]
    sel_pos_ = [sel.cursor.selectionStart(), sel.cursor.selectionEnd()]

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == cursor_pos
    assert sel_pos_ == sel_pos


def test_jump_mark_in_vline(vim_bot):
    """Jump to mark while in vline mode."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "ma")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(4)
    qtbot.keyClicks(cmd_line, "V'a")
    assert vim.vim_cmd.vim_status.vim_state == VimState.VLINE
    assert editor.textCursor().position() == 0


@pytest.mark.parametrize(
    "start_line, cmd_list, expected",
    [
        (0, ["V", "]]"], 2),
        (6, ["V", "]]"], 6),
        (9, ["V", "]]"], 9),
    ],
)
def test_next_python_definition_in_vline(vim_bot, start_line, cmd_list, expected):
    """Jump to next Python definition while in Vline mode."""
    _, _, editor, vim, qtbot = vim_bot

    text = (
        "import os\n\n"
        "def a():\n    pass\n\n"
        "class Foo:\n    def b(self):\n        pass\n\n"
        "def c():\n    pass\n"
    )
    editor.set_text(text)

    block = editor.document().findBlockByNumber(start_line)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(block.position())
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().blockNumber() == expected


@pytest.mark.parametrize(
    "start_line, cmd_list, expected",
    [
        (10, ["V", "[["], 9),
        (6, ["V", "[["], 5),
        (2, ["V", "[["], 2),
    ],
)
def test_prev_python_definition_in_vline(vim_bot, start_line, cmd_list, expected):
    """Jump to previous Python definition while in Vline mode."""
    _, _, editor, vim, qtbot = vim_bot

    text = (
        "import os\n\n"
        "def a():\n    pass\n\n"
        "class Foo:\n    def b(self):\n        pass\n\n"
        "def c():\n    pass\n"
    )
    editor.set_text(text)

    block = editor.document().findBlockByNumber(start_line)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(block.position())
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().blockNumber() == expected


@pytest.mark.parametrize(
    "start_line, cmd_list, expected",
    [
        (0, ["V", "}}"], 3),
        (3, ["V", "}}"], 6),
        (6, ["V", "}}"], 9),
        (9, ["V", "}}"], 12),
        (12, ["V", "}}"], 17),
        (17, ["V", "}}"], 20),
    ],
)
def test_next_python_block_in_vline(vim_bot, start_line, cmd_list, expected):
    """Jump to next Python block while in Vline mode."""
    _, _, editor, vim, qtbot = vim_bot

    text = (
        "def a():\n"
        "    pass\n\n"
        "if x:\n    pass\n\n"
        "for i in range(3):\n    pass\n\n"
        "while False:\n    pass\n\n"
        "try:\n    pass\nexcept Exception:\n    pass\n\n"
        "with open('file') as f:\n    pass\n\n"
        "class Foo:\n    def bar(self):\n        pass\n"
    )
    editor.set_text(text)

    block = editor.document().findBlockByNumber(start_line)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(block.position())
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().blockNumber() == expected


@pytest.mark.parametrize(
    "start_line, cmd_list, expected",
    [
        (20, ["V", "{{"], 17),
        (17, ["V", "{{"], 12),
        (12, ["V", "{{"], 9),
        (9, ["V", "{{"], 6),
        (6, ["V", "{{"], 3),
        (3, ["V", "{{"], 0),
    ],
)
def test_prev_python_block_in_vline(vim_bot, start_line, cmd_list, expected):
    """Jump to previous Python block while in Vline mode."""
    _, _, editor, vim, qtbot = vim_bot

    text = (
        "def a():\n"
        "    pass\n\n"
        "if x:\n    pass\n\n"
        "for i in range(3):\n    pass\n\n"
        "while False:\n    pass\n\n"
        "try:\n    pass\nexcept Exception:\n    pass\n\n"
        "with open('file') as f:\n    pass\n\n"
        "class Foo:\n    def bar(self):\n        pass\n"
    )
    editor.set_text(text)

    block = editor.document().findBlockByNumber(start_line)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(block.position())
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        qtbot.keyClicks(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert editor.textCursor().blockNumber() == expected



# -*- coding: utf-8 -*-
"""Base classes for OkVim command executors.

This module provides :class:`ExecutorBase` and :class:`ExecutorSubBase`, which
handle the parsing of command strings and delegation of actions to the
``MotionHelper`` and ``ActionHelper`` utilities. Executors interpret user
keystrokes, update the shared :class:`~spyder_okvim.vim.VimStatus` object and
decide whether submodes should be entered.
"""

# Standard Libraries
from typing import Any, NamedTuple

# Project Libraries
from spyder_okvim.utils.action_helpers import ActionHelper
from spyder_okvim.utils.motion_helpers import MotionHelper

FUNC_INFO = NamedTuple("FUNC_INFO", [("func", Any), ("has_arg", bool)])
RETURN_EXECUTOR_METHOD_INFO = NamedTuple(
    "RETURN_EXECUTOR_METHOD_INFO", [("sub_mode", Any), ("clear_command_line", bool)]
)


class ExecutorBase:
    """Baseclass Executor."""

    SYMBOLS_REPLACEMENT = {
        "!": "exclamation",
        "?": "question",
        "<": "less",
        ">": "greater",
        "|": "pipe",
        " ": "space",
        "\b": "backspace",
        "\r": "enter",
        "@": "at",
        "$": "dollar",
        "0": "zero",
        "^": "caret",
        '"': "quote",
        "'": "apostrophe",
        "`": "backtick",
        "%": "percent",
        "~": "tilde",
        ":": "colon",
        "/": "slash",
        ";": "semicolon",
        ",": "comma",
        ".": "dot",
        "[": "opensquarebracket",
        "]": "closesquarebracket",
        "*": "asterisk",
        "#": "sharp",
        "{": "openbrace",
        "}": "closebrace",
    }

    def __init__(self, vim_status):
        """Initialize the executor with the shared :class:`VimStatus`.

        Args:
            vim_status: :class:`spyder_okvim.vim.VimStatus` instance providing
                editor and cursor helpers.

        """
        self.vim_status = vim_status
        self.get_cursor = vim_status.get_cursor
        self.set_cursor = vim_status.set_cursor
        self.get_editor = vim_status.get_editor
        self.get_editorstack = vim_status.get_editorstack
        self.get_block_no_start_in_selection = (
            vim_status.get_block_no_start_in_selection
        )
        self.get_block_no_end_in_selection = vim_status.get_block_no_end_in_selection
        self.get_pos_start_in_selection = vim_status.get_pos_start_in_selection
        self.get_pos_end_in_selection = vim_status.get_pos_end_in_selection
        self.set_cursor_pos = vim_status.cursor.set_cursor_pos
        self.helper_motion = MotionHelper(vim_status)
        self.helper_action = ActionHelper(vim_status)

        self.has_zero_cmd = True
        self.pattern_cmd = None
        self.allow_leaderkey = True

    def update_input_cmd_info(self, num_str, cmd, input_txt):
        """Update input cmd to vim_status."""
        self.vim_status.input_cmd_prev.set(self.vim_status.input_cmd)
        self.vim_status.input_cmd.num_str = num_str
        self.vim_status.input_cmd.cmd = cmd

    def set_parent_info_to_submode(self, submode, num, num_str):
        """Set info to submode."""
        submode.parent_num = [num]
        submode.parent_num_str = [num_str]

    def __call__(self, txt: str) -> bool:
        """Parse txt and executor command.

        Returns:
            if return is True, Clear command line.

        """
        if self.has_zero_cmd and txt == "0":
            # Special case to simplify regexp
            key = "0"
            num_str = "1"
        else:
            match = self.pattern_cmd.match(txt)
            if not match:
                if txt.isdigit():
                    return False
                else:
                    self.vim_status.sub_mode = None
                    return True

            num_str, key = match.groups()

        num = int(num_str) if num_str else 1
        self.update_input_cmd_info(num_str, key, txt)

        key = self.SYMBOLS_REPLACEMENT.get(key, key)

        ret = None
        try:
            method = self.__getattribute__(key)
        except AttributeError:
            print("unknown key", key)
        else:
            ret = method(num=num, num_str=num_str)

        return self.process_return(ret)

    def process_return(self, ret: Any):
        """Process return of method."""
        if ret:
            self.vim_status.sub_mode = ret.sub_mode
            return ret.clear_command_line
        else:
            self.vim_status.sub_mode = None
            return True


class ExecutorSubBase(ExecutorBase):
    """Baseclass for submode of executor."""

    def __init__(self, vim_status):
        super().__init__(vim_status)

        self.has_zero_cmd = False
        self.parent_num = []
        self.parent_num_str = []
        self.func_list_deferred: list[FUNC_INFO] = []
        self.return_deferred: Any = None

    def set_parent_info_to_submode(self, submode, num, num_str):
        """Set parent and self into to submode."""
        submode.parent_num = self.parent_num.copy()
        submode.parent_num_str = self.parent_num_str.copy()

        submode.parent_num.append(num)
        submode.parent_num_str.append(num_str)

    def set_func_list_deferred(self, f_list: list[FUNC_INFO], ret: Any = None):
        """Set func list."""
        self.func_list_deferred = f_list
        self.return_deferred = ret

    def execute_func_deferred(self, arg=None):
        """Execute method passed in from previous executor."""
        for func_info in self.func_list_deferred:
            if func_info.has_arg:
                func_info.func(arg)
            else:
                func_info.func()
        return self.return_deferred

    def update_input_cmd_info(self, num_str, cmd, input_txt):
        """Add input cmd to vim_status."""
        self.vim_status.input_cmd.cmd += input_txt

    # ------------------------------------------------------------------
    # Common editor actions
    # ------------------------------------------------------------------
    def close_current_file(self) -> None:
        """Close the current file using Spyder's editor API."""
        self.vim_status.editor_widget.close_file()

    def save_current_file(self) -> None:
        """Save the active file through Spyder's editor plugin."""
        self.vim_status.editor_widget.save()

    def create_new_file(self) -> None:
        """Create a new file using Spyder's editor plugin."""
        self.vim_status.editor_widget.new()



"""Tests for JumpList utility"""

from spyder_okvim.utils.jump_list import JumpList, Jump


def test_push_and_navigation():
    jl = JumpList(max_items=5)
    jl.push("a.py", 1)
    jl.push("a.py", 2)
    jl.push("a.py", 3)

    assert jl.index == 3
    assert jl.jumps == [Jump("a.py", 1), Jump("a.py", 2), Jump("a.py", 3)]

    prev = jl.back()
    assert prev == Jump("a.py", 2)
    assert jl.index == 2

    nxt = jl.forward()
    assert nxt == Jump("a.py", 3)
    assert jl.index == 3


def test_duplicate_and_truncate():
    jl = JumpList(max_items=3)
    jl.push("a.py", 1)
    jl.push("a.py", 1)
    assert jl.jumps == [Jump("a.py", 1)]

    jl.push("b.py", 2)
    jl.push("c.py", 3)
    jl.push("d.py", 4)
    assert len(jl.jumps) == 3
    assert jl.jumps[0] == Jump("b.py", 2)

    jl.back()
    jl.push("e.py", 5)
    assert jl.jumps == [Jump("b.py", 2), Jump("c.py", 3), Jump("e.py", 5)]
    assert jl.index == 3



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License

"""Spyder okvim default configuration."""

# Third Party Libraries
from qtpy.QtCore import Qt

CONF_SECTION = "spyder_okvim"

KEYCODE2STR = {
    Qt.Key_Return: "\r",
    Qt.Key_Enter: "\r",
    Qt.Key_Space: " ",
    Qt.Key_Backspace: "\b",
}

CONF_DEFAULTS = [
    (
        CONF_SECTION,
        {
            "ignorecase": True,
            "smartcase": True,
            "use_sneak": True,
            "highlight_yank": True,
            "highlight_yank_duration": 400,
            "cursor_fg_color": "#000000",
            "cursor_bg_color": "#BBBBBB",
            "select_fg_color": "#A9B7C6",
            "select_bg_color": "#214283",
            "search_fg_color": "#A9B7C6",
            "search_bg_color": "#30652F",
            "yank_fg_color": "#17172d",
            "yank_bg_color": "#5cacee",
            "leader_key": "Space",
        },
    )
]

CONF_VERSION = "0.9"



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum


class MotionType(IntEnum):
    """Types of cursor motions."""

    BlockWise = 0
    LineWise = 1
    CharWise = 2
    CharWiseIncludingEnd = 3


@dataclass
class MotionInfo:
    """Information about a computed motion."""

    cursor_pos: int | None = None
    sel_start: int | None = None
    sel_end: int | None = None
    motion_type: MotionType = MotionType.LineWise



# -*- coding: utf-8 -*-
"""Utility functions that implement Vim-style editing actions."""

# Standard Libraries
import re

# Third Party Libraries
from qtpy.QtGui import QTextCursor

# Project Libraries
from spyder_okvim.utils.motion import MotionInfo, MotionType
from spyder_okvim.utils.motion_helpers import MotionHelper
from spyder_okvim.vim import VimState


def _get_block_texts(editor, start_block: int, end_block: int) -> list[str]:
    """Return the text of blocks between two numbers inclusive."""

    return [
        editor.document().findBlockByNumber(no).text()
        for no in range(start_block, end_block + 1)
    ]


class ActionHelper:
    """Perform editing actions driven by Vim commands."""

    def __init__(self, vim_status):
        """Initialize helper with the shared Vim status.

        Args:
            vim_status: Object managing editor and cursor state.
        """
        self.vim_status = vim_status
        self.get_editor = vim_status.get_editor
        self.get_cursor = vim_status.get_cursor
        self.set_cursor = vim_status.set_cursor
        self.set_cursor_pos = vim_status.cursor.set_cursor_pos

        self.get_block_no_start_in_selection = (
            vim_status.get_block_no_start_in_selection
        )
        self.get_block_no_end_in_selection = vim_status.get_block_no_end_in_selection
        self.get_pos_start_in_selection = vim_status.get_pos_start_in_selection
        self.get_pos_end_in_selection = vim_status.get_pos_end_in_selection
        self.helper_motion = MotionHelper(vim_status)

    def _get_block_range(
        self, pos_start: int, pos_end: int
    ) -> tuple[int, int, int, int]:
        """Return block numbers and adjusted positions for a range."""

        get_block = self.vim_status.cursor.get_block
        block_start, block_no_start = get_block(pos_start)
        block_end, block_no_end = get_block(pos_end)
        pos_start = block_start.position()
        pos_end = block_end.position() + block_end.length() - 1
        return block_no_start, block_no_end, pos_start, pos_end

    def _get_selection_range(
        self, motion_info: MotionInfo | None
    ) -> tuple[int | None, int | None, bool]:
        """Return selection start, end and linewise flag for an action."""

        if not self.vim_status.is_normal():
            start = self.get_pos_start_in_selection()
            end = self.get_pos_end_in_selection()
            is_linewise = self.vim_status.vim_state == VimState.VLINE
            return start, end, is_linewise

        if motion_info is None:
            return None, None, False

        if motion_info.motion_type == MotionType.BlockWise:
            return motion_info.sel_start, motion_info.sel_end, False
        if (
            motion_info.motion_type == MotionType.LineWise
            and motion_info.sel_start is not None
            and motion_info.sel_end is not None
        ):
            return motion_info.sel_start, motion_info.sel_end, True

        cursor = self.get_cursor()
        cur_pos = cursor.position()
        new_pos = motion_info.cursor_pos
        if new_pos is None:
            return None, None, False

        start, end = sorted([cur_pos, new_pos])
        if motion_info.motion_type == MotionType.CharWiseIncludingEnd:
            end += 1
        elif motion_info.motion_type == MotionType.LineWise:
            block_start, _ = self.vim_status.cursor.get_block(start)
            block_end, _ = self.vim_status.cursor.get_block(end)
            start = block_start.position()
            end = block_end.position() + block_end.length() - 1
            return start, end, True

        return start, end, False

    def join_lines(
        self, cursor_pos_start: int, block_no_start: int, block_no_end: int
    ) -> None:
        """Join lines between two blocks.

        Args:
            cursor_pos_start: Cursor position at the start of the first line.
            block_no_start: Number of the first block to join.
            block_no_end: Number of the last block to join.
        """
        self.vim_status.update_dot_cmd(connect_editor=False)

        editor = self.get_editor()
        cursor = editor.textCursor()
        n_block = editor.blockCount()

        if block_no_end >= n_block - 1:
            block_no_end = n_block - 1

        if block_no_start == block_no_end:
            return

        n_line = block_no_end - block_no_start + 1
        text_list = [""]
        cursor.setPosition(cursor_pos_start)
        for _ in range(n_line - 1):
            cursor.movePosition(QTextCursor.NextBlock)
            text = cursor.block().text().lstrip()
            if text:
                text_list.append(text)

        # Replace text
        cursor.setPosition(cursor_pos_start)
        cursor.movePosition(QTextCursor.EndOfLine)
        cursor.movePosition(QTextCursor.NextBlock, QTextCursor.KeepAnchor, n_line - 1)
        cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
        cursor.insertText(" ".join(text_list))

        # Move position of cursor
        cursor.movePosition(QTextCursor.EndOfBlock)
        cursor.movePosition(QTextCursor.Left, n=len(text_list[-1]) + 1)
        pos = cursor.position()

        self.set_cursor_pos(pos)

        editor.document_did_change()

    def replace_txt_with_ch(self, pos_start: int, pos_end: int, ch: str) -> None:
        """Replace selected text with a given character.

        Args:
            pos_start: Start position of the range to replace.
            pos_end: End position of the range to replace.
            ch: Character used for replacement.
        """
        self.vim_status.update_dot_cmd(connect_editor=False)

        if self.vim_status.is_normal() and pos_start == pos_end:
            self.vim_status.cursor.set_cursor_pos(pos_start)
            return

        cursor = self.get_cursor()
        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText().replace("\u2029", "\n")
        text_sub = re.sub(r".", ch, text)
        cursor.insertText(text_sub)

        if self.vim_status.is_normal():
            self.vim_status.cursor.set_cursor_pos(pos_end - 1)

        editor = self.get_editor()
        editor.document_did_change()

    def _add_surrounding(self, ch: str, text: str) -> str:
        """Return *text* wrapped by the given character."""
        prefix_dict = {
            "'": "'",
            '"': '"',
            "(": "( ",
            "{": "{ ",
            "[": "[ ",
            ")": "(",
            "}": "{",
            "]": "[",
        }

        suffix_dict = {
            "'": "'",
            '"': '"',
            "(": " )",
            "{": " }",
            "[": " ]",
            ")": ")",
            "}": "}",
            "]": "]",
        }

        return prefix_dict[ch] + text + suffix_dict[ch]

    def add_surrounding(self, pos_start: int, pos_end: int, ch: str) -> None:
        """Insert a surrounding character around the given range.

        Args:
            pos_start: Start position of the text to wrap.
            pos_end: End position of the text to wrap.
            ch: Surrounding character to add.
        """
        self.vim_status.update_dot_cmd(connect_editor=False)

        if self.vim_status.is_normal() and pos_start == pos_end:
            self.vim_status.cursor.set_cursor_pos(pos_start)
            return

        cursor = self.get_cursor()
        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText().replace("\u2029", "\n")
        text_sub = self._add_surrounding(ch, text)
        cursor.insertText(text_sub)

        editor = self.get_editor()
        editor.document_did_change()

    def _delete_surrounding(self, ch: str, text: str) -> str:
        """Return ``text`` without its surrounding character."""
        open_brackets = "([{"
        text_sub = text[1:-1]
        if ch in open_brackets:
            text_sub = text_sub.strip()
        return text_sub

    def delete_surrounding(self, ch: str) -> MotionInfo:
        """Remove surrounding characters around the current selection.

        Args:
            ch: Character used to locate the surroundings.

        Returns:
            Information about the motion that located the text.
        """
        self.vim_status.update_dot_cmd(connect_editor=False)

        if ch in "'\"":
            motion_info = self.helper_motion.i_quote(ch)
            if motion_info.sel_start and motion_info.sel_end:
                motion_info.sel_start -= 1
                motion_info.sel_end += 1
        else:
            motion_info = self.helper_motion.a_bracket(1, ch)

        if motion_info.sel_start is None:
            return motion_info

        cursor = self.get_cursor()
        cursor.setPosition(motion_info.sel_start)
        cursor.setPosition(motion_info.sel_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText().replace("\u2029", "\n")
        text_sub = self._delete_surrounding(ch, text)
        cursor.insertText(text_sub)

        editor = self.get_editor()
        editor.document_did_change()

        return motion_info

    def change_surrounding(self, ch_delete: str, ch_insert: str) -> MotionInfo:
        """Replace one surrounding character with another.

        Args:
            ch_delete: Character defining the surrounding to remove.
            ch_insert: Character to wrap around the text.

        Returns:
            Information about the motion that located the text.
        """
        self.vim_status.update_dot_cmd(connect_editor=False)

        if ch_delete in "'\"":
            motion_info = self.helper_motion.i_quote(ch_delete)
            if motion_info.sel_start and motion_info.sel_end:
                motion_info.sel_start -= 1
                motion_info.sel_end += 1
        else:
            motion_info = self.helper_motion.a_bracket(1, ch_delete)

        if motion_info.sel_start is None:
            return motion_info

        cursor = self.get_cursor()
        cursor.setPosition(motion_info.sel_start)
        cursor.setPosition(motion_info.sel_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText().replace("\u2029", "\n")
        text_sub = self._delete_surrounding(ch_delete, text)
        text_sub = self._add_surrounding(ch_insert, text_sub)
        cursor.insertText(text_sub)

        editor = self.get_editor()
        editor.document_did_change()

        return motion_info

    def handle_case(self, motion_info: MotionInfo, method) -> None:
        """Apply case transformation to text determined by ``motion_info``.

        Args:
            motion_info: Motion describing the range to transform.
            method: One of ``"swap"``, ``"lower"`` or ``"upper"``.
        """
        self.vim_status.update_dot_cmd(connect_editor=False)

        if not self.vim_status.is_normal():
            sel_start = self.get_pos_start_in_selection()
            sel_end = self.get_pos_end_in_selection()
            self._handle_case(sel_start, sel_end, method)
            self.vim_status.to_normal()
            self.vim_status.cursor.set_cursor_pos(sel_start)
        elif motion_info.motion_type == MotionType.BlockWise:
            self._handle_case(motion_info.sel_start, motion_info.sel_end, method)
            self.vim_status.cursor.set_cursor_pos(motion_info.sel_start)
        else:
            cursor = self.get_cursor()
            cursor_pos_cur = cursor.position()
            cursor_pos_new = motion_info.cursor_pos
            if cursor_pos_new is None:
                return
            pos_start, pos_end = sorted([cursor_pos_cur, cursor_pos_new])

            if motion_info.motion_type == MotionType.CharWise:
                self._handle_case(pos_start, pos_end, method)
            elif motion_info.motion_type == MotionType.CharWiseIncludingEnd:
                self._handle_case(pos_start, pos_end + 1, method)
            elif motion_info.motion_type == MotionType.LineWise:
                block_start, _ = self.vim_status.cursor.get_block(pos_start)
                block_end, _ = self.vim_status.cursor.get_block(pos_end)
                sel_start = block_start.position()
                sel_end = block_end.position() + block_end.length() - 1
                self._handle_case(sel_start, sel_end, method)

                # Exception case, guu, gUU, g~~
                if pos_start == pos_end:
                    pos_start = sel_start
            self.vim_status.cursor.set_cursor_pos(pos_start)

    def _handle_case(self, pos_start: int, pos_end: int, method) -> None:
        """Perform case transformation on the given range."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        cursor = self.get_cursor()

        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText()
        if method == "swap":
            cursor.insertText(text.swapcase())
        elif method == "lower":
            cursor.insertText(text.lower())
        else:
            cursor.insertText(text.upper())

        editor = self.get_editor()
        editor.document_did_change()

    def indent(self, motion_info: MotionInfo) -> None:
        """Shift the given lines to the right."""
        pos_start, pos_end, _ = self._get_selection_range(motion_info)
        if pos_start is None or pos_end is None:
            return
        self._indent(pos_start, pos_end)

    def _indent(self, pos_start: int, pos_end: int) -> None:
        """Indent lines between ``pos_start`` and ``pos_end``."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        editor = self.get_editor()
        cursor = self.get_cursor()

        block_no_start, block_no_end, pos_start, pos_end = self._get_block_range(
            pos_start, pos_end
        )

        text_list = _get_block_texts(editor, block_no_start, block_no_end)

        indent = self.vim_status.indent
        text_list_indent = []
        for text in text_list:
            if text:
                text_list_indent.append(indent + text)
            else:
                text_list_indent.append("")
        texts_indent = "\n".join(text_list_indent)

        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        cursor.insertText(texts_indent)

        block_start, _ = self.vim_status.cursor.get_block(pos_start)
        len_blank = len(block_start.text()) - len(block_start.text().lstrip())

        self.vim_status.cursor.set_cursor_pos(block_start.position() + len_blank)
        editor.document_did_change()

    def unindent(self, motion_info: MotionInfo) -> None:
        """Shift the given lines to the left."""
        pos_start, pos_end, _ = self._get_selection_range(motion_info)
        if pos_start is None or pos_end is None:
            return
        self._unindent(pos_start, pos_end)

    def _unindent(self, pos_start: int, pos_end: int) -> None:
        """Remove one level of indentation from the given range."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        editor = self.get_editor()
        cursor = self.get_cursor()

        block_no_start, block_no_end, pos_start, pos_end = self._get_block_range(
            pos_start, pos_end
        )

        text_list = _get_block_texts(editor, block_no_start, block_no_end)

        indent = self.vim_status.indent
        len_indent = len(indent)
        text_list_unindent = []
        for text in text_list:
            n_space = len(text) - len(text.lstrip())
            idx_discard = min([n_space, len_indent])
            text_list_unindent.append(text[idx_discard:])
        texts_unindent = "\n".join(text_list_unindent)

        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        cursor.insertText(texts_unindent)

        block_start, _ = self.vim_status.cursor.get_block(pos_start)
        len_blank = len(block_start.text()) - len(block_start.text().lstrip())

        self.vim_status.cursor.set_cursor_pos(block_start.position() + len_blank)
        editor.document_did_change()

    def yank(self, motion_info: MotionInfo, is_explicit: bool = False):
        """Copy text into the active register.

        Args:
            motion_info: Motion information for the yank range.
            is_explicit: When ``True`` the yank was explicitly requested.

        """
        if self.vim_status.is_normal():
            if motion_info.motion_type == MotionType.BlockWise:
                if motion_info.sel_start is None or motion_info.sel_end is None:
                    return None, None
            elif motion_info.cursor_pos is None:
                return None, None

        register_name = self.vim_status.get_register_name()
        sel_start, sel_end, is_linewise = self._get_selection_range(motion_info)
        if sel_start is None or sel_end is None:
            return None, None

        register_type = VimState.VLINE if is_linewise else VimState.NORMAL

        cursor = self.get_cursor()
        cursor.setPosition(sel_start)
        cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
        txt = cursor.selectedText().replace("\u2029", "\n")
        if register_type == VimState.VLINE:
            txt += "\n"

        self.vim_status.set_register(register_name, txt, register_type)
        if is_explicit is True and register_name == '"':
            self.vim_status.set_register("0", txt, register_type)

        # Set message
        doc = cursor.document()
        nb_start = doc.findBlock(sel_start).blockNumber()
        nb_end = doc.findBlock(sel_end).blockNumber()
        if nb_start != nb_end:
            self.vim_status.set_message(f"{nb_end - nb_start + 1} lines yanked")

        # highlight yank
        if is_explicit is True and self.vim_status.is_normal():
            if self.vim_status.running_dot_cmd is False:
                self.vim_status.cursor.highlight_yank(sel_start, sel_end)

        return sel_start, sel_end

    def _move_cursor_after_space(self, cursor):
        """Return ``cursor`` positioned after leading spaces."""
        txt = cursor.block().text()
        n_space = len(txt) - len(txt.lstrip())
        cursor.movePosition(QTextCursor.StartOfLine)
        cursor.movePosition(QTextCursor.Right, n=n_space)
        return cursor

    def paste_in_normal(self, num, is_lower):
        """Paste register content in normal mode."""
        reg = self.vim_status.get_register()
        self.vim_status.update_dot_cmd(connect_editor=False, register_name=reg.name)

        editor = self.get_editor()
        n_block_old = editor.blockCount()

        if reg.content:
            content = reg.content * num
        else:
            content = ""

        cursor = self.get_cursor()
        if reg.type != VimState.VLINE:
            if not cursor.atBlockEnd() and is_lower:
                cursor.movePosition(QTextCursor.Right)
            cursor.insertText(content)
            pos_end = cursor.selectionEnd()
            self.vim_status.to_normal()
            self.set_cursor_pos(pos_end - 1)
        elif is_lower is False:
            cursor.movePosition(QTextCursor.StartOfLine)
            block_number_old = cursor.block().blockNumber()
            cursor.insertText(content)
            block = self.get_editor().document().findBlockByNumber(block_number_old)
            cursor.setPosition(block.position())
            cursor = self._move_cursor_after_space(cursor)
            self.set_cursor_pos(cursor.position())
        else:
            cursor.movePosition(QTextCursor.EndOfLine)
            cursor_pos_old = cursor.position()
            if cursor.atEnd():
                cursor.insertText("\n" + content[:-1])
            else:
                cursor.movePosition(QTextCursor.Right)
                cursor.insertText(content)
            cursor.setPosition(cursor_pos_old)
            cursor.movePosition(QTextCursor.Down)
            cursor = self._move_cursor_after_space(cursor)
            self.set_cursor_pos(cursor.position())

        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            self.vim_status.set_message(f"{n_block_new - n_block_old} more lines")

        editor = self.get_editor()
        editor.document_did_change()

    def paste_in_visual(self, num):
        """Paste over the current visual selection."""
        editor = self.get_editor()
        reg = self.vim_status.get_register()
        self.vim_status.update_dot_cmd(connect_editor=False, register_name=reg.name)

        editor = self.get_editor()
        n_block_old = editor.blockCount()

        if reg.content:
            content = reg.content * num
            if reg.type == VimState.VLINE:
                content = "\n" + content
        else:
            content = ""

        cursor = self.get_cursor()
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()
        block_no_start = self.get_block_no_start_in_selection()

        cursor.setPosition(sel_start)
        cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
        cursor.insertText(content)

        if reg.type == VimState.VLINE:
            block = editor.document().findBlockByNumber(block_no_start + 1)
            cursor_pos_new = block.position()
        else:
            if "\n" in content:
                cursor_pos_new = sel_start
            else:
                cursor_pos_new = sel_start + len(content) - 1

        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            self.vim_status.set_message(f"{n_block_new - n_block_old} more lines")

        editor.document_did_change()

        self.vim_status.to_normal()
        self.vim_status.cursor.set_cursor_pos(cursor_pos_new)

    def paste_in_vline(self, num):
        """Paste register content in visual line mode."""
        editor = self.get_editor()
        reg = self.vim_status.get_register()
        self.vim_status.update_dot_cmd(connect_editor=False, register_name=reg.name)

        editor = self.get_editor()
        n_block_old = editor.blockCount()

        if reg.content:
            if reg.type == VimState.VLINE:
                content = reg.content * num
                content = content[:-1]
            else:
                content = reg.content
                if num > 1:
                    content = (reg.content + "\n") * num
                    content = content[:-1]
        else:
            content = ""

        cursor = self.get_cursor()
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()

        cursor.setPosition(sel_start)
        cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
        cursor.insertText(content)

        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            self.vim_status.set_message(f"{n_block_new - n_block_old} more lines")

        editor.document_did_change()

        self.vim_status.to_normal()
        self.vim_status.cursor.set_cursor_pos(sel_start)

    def delete(self, motion_info: MotionInfo, is_insert, replace_txt=""):
        """Delete text defined by ``motion_info``.

        Args:
            motion_info: Motion describing the range to delete.
            is_insert: If ``True`` delete while in insert mode.
            replace_txt: Optional replacement text.
        """
        reg = self.vim_status.get_register()
        if is_insert:
            self.vim_status.update_dot_cmd(connect_editor=True, register_name=reg.name)
        else:
            self.vim_status.update_dot_cmd(connect_editor=False, register_name=reg.name)

        if self.vim_status.is_normal():
            if motion_info.motion_type == MotionType.BlockWise:
                if motion_info.sel_start is None or motion_info.sel_end is None:
                    return
            elif motion_info.cursor_pos is None:
                return

        cursor = self.get_cursor()
        editor = self.get_editor()
        n_block_old = editor.blockCount()
        sel_start, sel_end, is_linewise = self._get_selection_range(motion_info)
        if sel_start is None or sel_end is None:
            return

        if is_linewise:
            n_block = editor.blockCount()
            get_block = self.vim_status.cursor.get_block
            block_start, block_no_start = get_block(sel_start)
            block_end, block_no_end = get_block(sel_end)
            if is_insert:
                sel_start = block_start.position()
                sel_end = block_end.position() + block_end.length() - 1
            else:
                if block_no_start == 0 and block_no_end == n_block - 1:
                    sel_start = block_start.position()
                    sel_end = block_end.position() + block_end.length() - 1
                elif block_no_start == 0:
                    sel_start = block_start.position()
                    sel_end = block_end.position() + block_end.length()
                else:
                    sel_start = block_start.position() - 1
                    sel_end = block_end.position() + block_end.length() - 1

        cursor.setPosition(sel_start)
        cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
        cursor.insertText(replace_txt)

        # Set message
        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            self.vim_status.set_message(f"{n_block_old - n_block_new} fewer lines")

        # Set cursor pos in normal
        if self.vim_status.is_normal():
            cursor_pos_new = sel_start
            if is_linewise and is_insert is False:
                n_block = editor.blockCount()
                block_no = min([block_no_start, n_block - 1])
                block = editor.document().findBlockByNumber(block_no)
                txt = block.text()
                cursor_pos_new = block.position() + len(txt) - len(txt.lstrip())
                cursor.setPosition(cursor_pos_new)

            if is_insert:
                self.set_cursor_pos(cursor_pos_new + len(replace_txt))
            else:
                self.vim_status.cursor.set_cursor_pos_without_end(cursor_pos_new)

        editor.document_did_change()

    def toggle_comment(self, motion_info: MotionInfo):
        """Toggle comments for the selected lines."""
        self.vim_status.update_dot_cmd(connect_editor=False)
        editor = self.get_editor()
        pos_start, pos_end, _ = self._get_selection_range(motion_info)
        if pos_start is None or pos_end is None:
            return

        block_no_start, _, pos_start, pos_end = self._get_block_range(
            pos_start, pos_end
        )

        new_cursor = editor.textCursor()
        new_cursor.setPosition(pos_start, QTextCursor.MoveAnchor)
        new_cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        editor.setTextCursor(new_cursor)
        editor.toggle_comment()

        new_cursor = editor.textCursor()
        block = editor.document().findBlockByNumber(block_no_start)
        pos_start = block.position()
        new_cursor.setPosition(pos_start, QTextCursor.MoveAnchor)
        editor.setTextCursor(new_cursor)

        self.vim_status.to_normal()
        self.vim_status.cursor.set_cursor_pos(pos_start)

        editor.document_did_change()



"""Manage the jump list for navigating positions."""

from dataclasses import dataclass


@dataclass
class Jump:
    file: str
    pos: int


class JumpList:
    """Simple jump list implementation."""

    def __init__(self, max_items: int = 100) -> None:
        self.max_items = max_items
        self.jumps: list[Jump] = []
        self.index: int = 0

    def push(self, file: str, pos: int) -> None:
        """Add a new jump location."""
        if self.index < len(self.jumps):
            self.jumps = self.jumps[: self.index]
        if self.jumps and self.jumps[-1].file == file and self.jumps[-1].pos == pos:
            return
        self.jumps.append(Jump(file, pos))
        if len(self.jumps) > self.max_items:
            self.jumps.pop(0)
        self.index = len(self.jumps)

    def back(self) -> Jump | None:
        """Return previous jump and update index."""
        if self.index <= 1:
            return None
        self.index -= 1
        return self.jumps[self.index - 1]

    def forward(self) -> Jump | None:
        """Return next jump and update index."""
        if self.index >= len(self.jumps):
            return None
        jump = self.jumps[self.index]
        self.index += 1
        return jump

    def pop_last(self) -> None:
        """Remove the most recently added jump if present."""
        if not self.jumps:
            return
        self.jumps.pop()
        if self.index > len(self.jumps):
            self.index = len(self.jumps)
